@@ -21,6 +21,7 @@ import functools
 import platform
 import sys
 
+from nova import context as nova_context
 from nova import exception
 from nova import image
 from nova.virt import driver
@@ -96,10 +97,14 @@ exception_conversion_map = {
 class HyperVDriver(driver.ComputeDriver):
     capabilities = {
         "has_imagecache": True,
-        "supports_recreate": False,
+        "supports_recreate": True,
         "supports_migrate_to_same_host": False,
         "supports_attach_interface": True,
         "supports_device_tagging": True,
+        "supports_tagged_attach_interface": True,
+        "supports_tagged_attach_volume": True,
+        "supports_extend_volume": True,
+        "supports_multiattach": False,
     }
 
     def __init__(self, virtapi):
@@ -122,7 +127,8 @@ class HyperVDriver(driver.ComputeDriver):
         self._pathutils = pathutils.PathUtils()
 
     def _check_minimum_windows_version(self):
-        if not utilsfactory.get_hostutils().check_min_windows_version(6, 2):
+        hostutils = utilsfactory.get_hostutils()
+        if not hostutils.check_min_windows_version(6, 2):
             # the version is of Windows is older than Windows Server 2012 R2.
             # Log an error, letting users know that this version is not
             # supported any longer.
@@ -131,6 +137,12 @@ class HyperVDriver(driver.ComputeDriver):
                       'Server 2012). The support for this version of '
                       'Windows has been removed in Mitaka.')
             raise exception.HypervisorTooOld(version='6.2')
+        elif not hostutils.check_min_windows_version(6, 3):
+            # TODO(claudiub): replace the warning with an exception in Rocky.
+            LOG.warning('You are running nova-compute on Windows / Hyper-V '
+                        'Server 2012. The support for this version of Windows '
+                        'has been deprecated In Queens, and will be removed '
+                        'in Rocky.')
 
     @property
     def need_legacy_block_device_info(self):
@@ -155,7 +167,8 @@ class HyperVDriver(driver.ComputeDriver):
         return self._vmops.estimate_instance_overhead(instance_info)
 
     def spawn(self, context, instance, image_meta, injected_files,
-              admin_password, network_info=None, block_device_info=None):
+              admin_password, allocations, network_info=None,
+              block_device_info=None):
         image_meta = self._recreate_image_meta(context, instance, image_meta)
         self._vmops.spawn(context, instance, image_meta, injected_files,
                           admin_password, network_info, block_device_info)
@@ -179,13 +192,23 @@ class HyperVDriver(driver.ComputeDriver):
 
     def attach_volume(self, context, connection_info, instance, mountpoint,
                       disk_bus=None, device_type=None, encryption=None):
-        return self._volumeops.attach_volume(connection_info,
-                                             instance.name)
+        self._volumeops.attach_volume(context,
+                                      connection_info,
+                                      instance,
+                                      update_device_metadata=True)
 
     def detach_volume(self, connection_info, instance, mountpoint,
                       encryption=None):
-        return self._volumeops.detach_volume(connection_info,
-                                             instance.name)
+        context = nova_context.get_admin_context()
+        # The nova compute manager only updates the device metadata in
+        # case of tagged devices. We're including untagged devices as well.
+        self._volumeops.detach_volume(context,
+                                      connection_info,
+                                      instance,
+                                      update_device_metadata=True)
+
+    def extend_volume(self, connection_info, instance):
+        self._volumeops.extend_volume(connection_info)
 
     def get_volume_connector(self, instance):
         return self._volumeops.get_volume_connector()
@@ -366,9 +389,10 @@ class HyperVDriver(driver.ComputeDriver):
         self._imagecache.update(context, all_instances)
 
     def attach_interface(self, context, instance, image_meta, vif):
-        return self._vmops.attach_interface(instance, vif)
+        self._vmops.attach_interface(context, instance, vif)
 
     def detach_interface(self, context, instance, vif):
+        # The device metadata gets updated outside the driver.
         return self._vmops.detach_interface(instance, vif)
 
     def rescue(self, context, instance, network_info, image_meta,
@@ -400,3 +424,21 @@ class HyperVDriver(driver.ComputeDriver):
                                         in instance.system_metadata.items()}
         image_meta["id"] = image_ref
         return image_meta
+
+    def check_instance_shared_storage_local(self, context, instance):
+        """Check if instance files located on shared storage.
+
+        This runs check on the destination host, and then calls
+        back to the source host to check the results.
+
+        :param context: security context
+        :param instance: nova.objects.instance.Instance object
+        :returns: A dict containing the tempfile info.
+        """
+        return self._pathutils.check_instance_shared_storage_local(instance)
+
+    def check_instance_shared_storage_remote(self, context, data):
+        return self._pathutils.check_instance_shared_storage_remote(data)
+
+    def check_instance_shared_storage_cleanup(self, context, data):
+        return self._pathutils.check_instance_shared_storage_cleanup(data)
