@@ -72,38 +72,74 @@ class VolumeOps(object):
 
     def __init__(self):
         self._volume_api = cinder.API()
+        self._vmops_prop = None
+        self._block_dev_man_prop = None
 
         self._vmutils = utilsfactory.get_vmutils()
         self._default_root_device = 'vda'
+
+        self._load_volume_drivers()
+
+    def _load_volume_drivers(self):
         self.volume_drivers = {
             constants.STORAGE_PROTOCOL_SMBFS: SMBFSVolumeDriver(),
             constants.STORAGE_PROTOCOL_ISCSI: ISCSIVolumeDriver(),
             constants.STORAGE_PROTOCOL_FC: FCVolumeDriver()}
 
+    @property
+    def _vmops(self):
+        # We have to avoid a circular dependency.
+        if not self._vmops_prop:
+            self._vmops_prop = importutils.import_class(
+                'compute_hyperv.nova.vmops.VMOps')()
+        return self._vmops_prop
+
+    @property
+    def _block_dev_man(self):
+        if not self._block_dev_man_prop:
+            self._block_dev_man_prop = importutils.import_class(
+                'compute_hyperv.nova.block_device_manager.'
+                'BlockDeviceInfoManager')()
+        return self._block_dev_man_prop
+
     def _get_volume_driver(self, connection_info):
         driver_type = connection_info.get('driver_volume_type')
         if driver_type not in self.volume_drivers:
             raise exception.VolumeDriverNotFound(driver_type=driver_type)
         return self.volume_drivers[driver_type]
 
-    def attach_volumes(self, volumes, instance_name):
+    def validate_host_configuration(self):
+        for protocol, volume_driver in self.volume_drivers.items():
+            try:
+                volume_driver.validate_host_configuration()
+            except exception.ValidationError as ex:
+                LOG.warning(
+                    "Volume driver %(protocol)s reported a validation "
+                    "error. Attaching such volumes will probably fail. "
+                    "Error message: %(err_msg)s.",
+                    dict(protocol=protocol, err_msg=ex.message))
+
+    def attach_volumes(self, context, volumes, instance):
         for vol in volumes:
-            self.attach_volume(vol['connection_info'], instance_name)
+            self.attach_volume(context, vol['connection_info'], instance)
 
     def disconnect_volumes(self, block_device_info):
         mapping = driver.block_device_info_get_mapping(block_device_info)
         for vol in mapping:
             self.disconnect_volume(vol['connection_info'])
 
-    def attach_volume(self, connection_info, instance_name,
-                      disk_bus=constants.CTRL_TYPE_SCSI):
+    def attach_volume(self, context, connection_info, instance,
+                      disk_bus=constants.CTRL_TYPE_SCSI,
+                      update_device_metadata=False):
         tries_left = CONF.hyperv.volume_attach_retry_count + 1
 
         while tries_left:
             try:
-                self._attach_volume(connection_info,
-                                    instance_name,
-                                    disk_bus)
+                self._attach_volume(context,
+                                    connection_info,
+                                    instance,
+                                    disk_bus,
+                                    update_device_metadata)
                 break
             except Exception as ex:
                 tries_left -= 1
@@ -113,9 +149,15 @@ class VolumeOps(object):
                         "to instance %(instance_name)s. ",
                         {'connection_info': strutils.mask_dict_password(
                              connection_info),
-                         'instance_name': instance_name})
+                         'instance_name': instance.name})
 
-                    self.disconnect_volume(connection_info)
+                    # We're requesting a detach as the disk may have
+                    # been attached to the instance but one of the
+                    # post-attach operations failed.
+                    self.detach_volume(context,
+                                       connection_info,
+                                       instance,
+                                       update_device_metadata)
                     raise exception.VolumeAttachFailed(
                         volume_id=connection_info['serial'],
                         reason=ex)
@@ -126,22 +168,32 @@ class VolumeOps(object):
                         "Tries left: %(tries_left)s.",
                         {'connection_info': strutils.mask_dict_password(
                              connection_info),
-                         'instance_name': instance_name,
+                         'instance_name': instance.name,
                          'tries_left': tries_left})
 
                     time.sleep(CONF.hyperv.volume_attach_retry_interval)
 
-    def _attach_volume(self, connection_info, instance_name,
-                       disk_bus=constants.CTRL_TYPE_SCSI):
+    def _attach_volume(self, context, connection_info, instance,
+                       disk_bus=constants.CTRL_TYPE_SCSI,
+                       update_device_metadata=False):
         LOG.debug(
             "Attaching volume: %(connection_info)s to %(instance_name)s",
             {'connection_info': strutils.mask_dict_password(connection_info),
-             'instance_name': instance_name})
+             'instance_name': instance.name})
         volume_driver = self._get_volume_driver(connection_info)
         volume_driver.attach_volume(connection_info,
-                                    instance_name,
+                                    instance.name,
                                     disk_bus)
 
+        if update_device_metadata:
+            # When attaching volumes to already existing instances,
+            # the connection info passed to the driver is not saved
+            # yet within the BDM table.
+            self._block_dev_man.set_volume_bdm_connection_info(
+                context, instance, connection_info)
+            self._vmops.update_device_metadata(
+                context, instance)
+
         qos_specs = connection_info['data'].get('qos_specs') or {}
         if qos_specs:
             volume_driver.set_disk_qos_specs(connection_info,
@@ -151,16 +203,20 @@ class VolumeOps(object):
         volume_driver = self._get_volume_driver(connection_info)
         volume_driver.disconnect_volume(connection_info)
 
-    def detach_volume(self, connection_info, instance_name):
+    def detach_volume(self, context, connection_info, instance,
+                      update_device_metadata=False):
         LOG.debug("Detaching volume: %(connection_info)s "
                   "from %(instance_name)s",
                   {'connection_info': strutils.mask_dict_password(
                       connection_info),
-                   'instance_name': instance_name})
+                   'instance_name': instance.name})
         volume_driver = self._get_volume_driver(connection_info)
-        volume_driver.detach_volume(connection_info, instance_name)
+        volume_driver.detach_volume(connection_info, instance.name)
         volume_driver.disconnect_volume(connection_info)
 
+        if update_device_metadata:
+            self._vmops.update_device_metadata(context, instance)
+
     def fix_instance_volume_disk_paths(self, instance_name, block_device_info):
         # Mapping containing the current disk paths for each volume.
         actual_disk_mapping = self.get_disk_path_mapping(block_device_info)
@@ -246,6 +302,7 @@ class VolumeOps(object):
                    "create_info": create_info})
         snapshot_id = create_info['snapshot_id']
 
+        snapshot_failed = False
         try:
             instance.task_state = task_states.IMAGE_SNAPSHOT_PENDING
             instance.save(expected_task_state=[None])
@@ -262,22 +319,23 @@ class VolumeOps(object):
             # The volume driver is expected to
             # update the connection info.
             driver_bdm.save()
-
-            self._volume_api.update_snapshot_status(
-                context, snapshot_id, 'creating')
         except Exception:
             with excutils.save_and_reraise_exception():
+                snapshot_failed = True
+
                 err_msg = ('Error occurred while snapshotting volume. '
                            'sending error status to Cinder.')
                 LOG.exception(err_msg,
                               instance=instance)
-                self._volume_api.update_snapshot_status(
-                    context, snapshot_id, 'error')
         finally:
             instance.task_state = None
             instance.save(
                 expected_task_state=[task_states.IMAGE_SNAPSHOT_PENDING])
 
+            snapshot_status = 'error' if snapshot_failed else 'creating'
+            self._volume_api.update_snapshot_status(
+                context, snapshot_id, snapshot_status)
+
     @volume_snapshot_lock
     def volume_snapshot_delete(self, context, instance, volume_id,
                                snapshot_id, delete_info):
@@ -287,6 +345,7 @@ class VolumeOps(object):
                    "instance_name": instance.name,
                    "delete_info": delete_info})
 
+        snapshot_delete_failed = False
         try:
             instance.task_state = task_states.IMAGE_SNAPSHOT_PENDING
             instance.save(expected_task_state=[None])
@@ -303,22 +362,32 @@ class VolumeOps(object):
             # The volume driver is expected to
             # update the connection info.
             driver_bdm.save()
-
-            self._volume_api.update_snapshot_status(
-                context, snapshot_id, 'deleting')
         except Exception:
             with excutils.save_and_reraise_exception():
+                snapshot_delete_failed = True
+
                 err_msg = ('Error occurred while deleting volume '
                            'snapshot. Sending error status to Cinder.')
                 LOG.exception(err_msg,
                               instance=instance)
-                self._volume_api.update_snapshot_status(
-                    context, snapshot_id, 'error_deleting')
         finally:
             instance.task_state = None
             instance.save(
                 expected_task_state=[task_states.IMAGE_SNAPSHOT_PENDING])
 
+            snapshot_status = ('error_deleting'
+                               if snapshot_delete_failed else 'deleting')
+            self._volume_api.update_snapshot_status(
+                context, snapshot_id, snapshot_status)
+
+    def get_disk_attachment_info(self, connection_info):
+        volume_driver = self._get_volume_driver(connection_info)
+        return volume_driver.get_disk_attachment_info(connection_info)
+
+    def extend_volume(self, connection_info):
+        volume_driver = self._get_volume_driver(connection_info)
+        return volume_driver.extend_volume(connection_info)
+
 
 class BaseVolumeDriver(object):
     _is_block_dev = True
@@ -361,6 +430,10 @@ class BaseVolumeDriver(object):
 
         return self._get_disk_res_path(disk_paths[0])
 
+    def validate_host_configuration(self):
+        if self._is_block_dev:
+            self._check_san_policy()
+
     def _get_disk_res_path(self, disk_path):
         if self._is_block_dev:
             # We need the Msvm_DiskDrive resource path as this
@@ -378,8 +451,22 @@ class BaseVolumeDriver(object):
             raise exception.DiskNotFound(err_msg)
         return disk_res_path
 
+    def _check_san_policy(self):
+        disk_policy = self._diskutils.get_new_disk_policy()
+
+        accepted_policies = [os_win_const.DISK_POLICY_OFFLINE_SHARED,
+                             os_win_const.DISK_POLICY_OFFLINE_ALL]
+
+        if disk_policy not in accepted_policies:
+            err_msg = _("Invalid SAN policy. The SAN policy "
+                        "must be set to 'Offline Shared' or 'Offline All' "
+                        "in order to attach passthrough disks to instances.")
+            raise exception.ValidationError(message=err_msg)
+
     def attach_volume(self, connection_info, instance_name,
                       disk_bus=constants.CTRL_TYPE_SCSI):
+        self.validate_host_configuration()
+
         dev_info = self.connect_volume(connection_info)
 
         serial = connection_info['serial']
@@ -456,6 +543,24 @@ class BaseVolumeDriver(object):
     def delete_snapshot(self, connection_info, instance, delete_info):
         raise NotImplementedError()
 
+    def get_disk_attachment_info(self, connection_info):
+        if self._is_block_dev:
+            disk_path = None
+            serial = connection_info['serial']
+        else:
+            disk_path = self.get_disk_resource_path(connection_info)
+            serial = None
+
+        return self._vmutils.get_disk_attachment_info(
+            disk_path,
+            is_physical=self._is_block_dev,
+            serial=serial)
+
+    def extend_volume(self, connection_info):
+        # We're not actually extending the volume, we're just
+        # refreshing cached information about an already extended volume.
+        self._connector.extend_volume(connection_info['data'])
+
 
 class ISCSIVolumeDriver(BaseVolumeDriver):
     _is_block_dev = True
